// Loki.java stand: 20020707
/**
 * Loki: unser spieler. wird entweder als einzelner Spieler direkt vom
 *       Prompt gestartet, oder als komplette Mannschaft vom RoboClient
 *
 * uebernimmt:
 * <ul><li>Verwaltung von Kommunikation, Weltbild und Strategie</ul>
 *
 * @author <a href="mailto:rmathes@uni-mannheim.de">Rene M</a>
 * @author <a href="mailto:dabrecht@uni-mannheim.de">David Brechtel</a>
 * @version 0.0.1
 * @see <code>RoboClient</code>
 * @see <code>kommunikation</code>
 * @see <code>Weltbild</code>
 * @see <code>Strategie</code>
 * @see <code>Visualisierung</code>
 */
public class Loki extends Thread {
// versionsnummer
  private static final String VERSION = "0.0.1";

  /**
   * diese Objekte werden zum zwischenspeichern und zur weitergabe ans weltbild
   * verwendet.
   */
  private kommunikation kom;
  private Weltbild world;
  private Strategie strategie;
  private Visualisierung visor;
  private hearObject ho;
  private seeObject[] so;
  private sense_bodyObject s_bo;
  private WbObjekt[] wo;
  private initObject io;
  private errorObject eo;

  private static String teamName;
  private boolean goalie;
  private boolean visible;
  private boolean stillRunning;

  /**
   * Loki( ... ) - der Konstruktor unseres Spielers
   * @param <code>String</code> serverName : der name des soccerservers
   * @param <code>int</code> serverPort : der port an den connected werden soll
   * @param <code>String</code> teamName : der name der mannschaft
   * @param <code>boolean</code> goalie : ist der spieler ein torhueter?
   * @param <code>visual</code> visual : soll die visualisierung aktiv sein?
   */
  public Loki( String serverName, int serverPort, String teamname, boolean goalie, boolean visual ) {
    setName( "Loki" );
    kom = new kommunikation( serverName, serverPort, teamname );
    this.teamName = teamname;
    world = new Weltbild(teamName);
    this.goalie = goalie;
    /**
     * wenn der spieler torhueter sein soll wird eine andere strategie
     * instanziert, ansonsten wird die normale spielerstrategie geladen.
     */
    if( goalie ) {
      GoalieStrategie strategie = new GoalieStrategie();
      this.strategie = strategie;
    } else {
      SpielerStrategie strategie = new SpielerStrategie();
      this.strategie = strategie;
    }
    /**
     * die visualisierung wird nur instanziert, wenn benoetigt.
     */
    visible = false;
    if( visual ) {
      visible = true;
      visor = new Visualisierung();
    }
    kom.init( goalie );

  }

  /**
   * void run() - die hauptroutine des threads, hier verwaltet Loki die
   * einzelnen routinen.
   */
  public void run() {
    stillRunning = true;
    while( stillRunning ) {
      kom.getObjects();
      if( kom.neuesHearObject() ) {
//        System.out.println( "Loki: in hearObject" );
        // hearObject abfragen
        ho = kom.getHearObject();
        // aufruf des weltbildes mit den korrekten daten...
        wo = world.aktualisiere( ho );
        //System.out.println("Loki: Weltbild aktualisiert ");
        stillRunning = strategie.handle( kom, wo );
      }
      if( kom.neuesSeeObject() ) {
        //System.out.println( "Loki: in seeObject" );
        // seeObject[] abfragen
        so = kom.getSeeObject();
        // aufruf des weltbildes mit den korrekten daten...
        // auchtung bei der parameter uebergabe hier. ist seeObject[]
        wo = world.aktualisiere( so );
        //System.out.println("Loki: Weltbild aktualisiert ");
        if( visible ) {
            visor.zeichne( wo );
        }
        stillRunning = strategie.handle( kom, wo );
     }
      if( kom.neuesSense_bodyObject() ) {
        //System.out.println( "Loki: in sense_bodyObject" );
        // sense_bodyObject abfragen
        s_bo = kom.getSense_bodyObject();
        // aufruf des weltbildes mit den korrekten daten...
        wo = world.aktualisiere( s_bo );
        //System.out.println("Loki: Weltbild aktualisiert ");
        // stillRunning = strategie.handle( kom, wo );
      }
      if( kom.neuesInitObject() ) {
        //System.out.println( "Loki: in initObject" );
        // initObject abfragen
        io = kom.getInitObject();
        // aufruf des weltbildes mit den korrekten daten...
        wo = world.aktualisiere( io );
        //System.out.println("Loki: Weltbild aktualisiert ");
        // stillRunning = strategie.handle( kom, wo );
      }
/** schlafen legen, ich hoffe das ist nicht schon zu lange... */
      try { sleep( 20 ); }
      catch( InterruptedException IE ) { }
    }
    System.out.println( "Loki: beende den Thread." );
  }

/**
 * displaySyntax() - wird bei angabe von falschen parametern aufgerufen
 */
  private static void displaySyntax() {
    System.out.println( "Syntax:\nLoki <teamName> [serverName|serverIp] [serverPort] [-visual]\n" );
    System.out.println( "der Name des Teams muss angegeben werden, alle anderen Parameter sind optional." );
  }
// ende von displaySyntax

/**
 * in der main() methode wird lediglich die kommandozeile ueberprueft, und, wenn korrekt Loki instanziert.
 * diese methode wird nur ausgefuehrt, wenn Loki vom prompt aufgerufen wird.
 */
  public static void main( String[] args ){
    String serverName = "localhost";
    int serverPort = 6000;
    teamName = "";
    boolean visible = false;
    System.out.println( "Loki - RoboCup Soccer Client version " + VERSION + " ; Gruppe 15\n\n" );
    /** parsing der kommandozeilen parameter */
    if( args.length == 0) {
      displaySyntax();
      System.exit( 1 );
    } else {
      teamName = args[ 0 ];
      if( args.length >1 ) {
        if( args[ 1 ].startsWith( "-visual" ) ) {
       	  visible = true;
       	} else {
       	  serverName = args[ 1 ];
       	}
      }
      if( args.length >2 ) {
        serverName = args[ 1 ];
       	if( args[ 2 ].startsWith( "-visual" ) ) {
       	  visible = true;
       	} else {
          try {
            serverPort = Integer.parseInt( args[ 2 ] );
          } catch( NumberFormatException NFE ) {
            System.out.println( "Parameter 3 nicht erkannt, sollte serverPort sein, ist: "+args[ 2 ]+"\n\n" );
            displaySyntax();
            System.exit( 1 );
          }
       	}
      }
      if( args.length >3 ) {
        try {
          serverPort = Integer.parseInt( args[ 2 ] );
        } catch( NumberFormatException NFE ) {
          System.out.println( "Parameter 3 nicht erkannt, sollte serverPort sein, ist: "+args[ 2 ]+"\n\n" );
          displaySyntax();
          System.exit( 1 );
        }
        if( args[ 3 ].startsWith( "-visual" ) ) {
       	  visible = true;
       	}
      }
    }
    /** ende des parsing */
    /** und Loki instanzieren...*/
    Loki ichbins = new Loki( serverName, serverPort, teamName, false, visible );
    ichbins.start();
    while( ichbins.isAlive() ) {
    // nix
    }
    System.out.println( "Loki beendet..." );
  }
  // ende der main methode
}
// ende der klasse Loki
/**
 * -----
 * ...
 * Loki ist schmuck und schn von Gestalt, aber bs von Gemt und sehr unbestndig.
 * Er bertrifft alle andern in Schlauheit und in jeder Art von Betrug.
 * ...
 * ( Stange: jngere Edda Str. 33, S. 282)
 * -----
 */
