import java.util.*;

/**
 * Die <code>Weltbild</code> Klasse dient dazu, aus den von
 * der <code>kommunikation</code> gelieferten Werten absolute
 * und von der <code>Strategie</code> brauchbare Daten ber
 * <ul>
 * <li>den Spieler (<code>WbSpielerObjekt</code>),
 * <li>das Spiel (<code>WbSpielObjekt</code>),
 * <li>den Ball (<code>WbBallObjekt</code>) und
 * <li>die Mitspieler (<code>WbAndereSpielerObjekt</code>)
 * </ul>
 * zu ermitteln und an die aufrufende Methode zurckzugeben.
 * Initialisiert wird das <code>Weltbild</code> beim Starten
 * des Spielers <code>Loki</code>.
 * Solange dieser Spieler existiert und die von der
 * <code>kommunikation</code> empfangenen Daten an das
 * <code>Weltbild</code> weitergibt, geschieht das, indem
 * eine der <code>aktualisiere</code>-Methoden aufgerufen wird.
 * Diese verarbeitet die Daten, mit denen sie aufgerufen wurde
 * und gibt die errechneten Daten an die aufrufende Methode
 * zurck.
 *
 *
 * @author <a href="mailto:dabrecht@uni-mannheim.de">David Brechtel</a>
 * @version 1.0
 *
 * @see Loki
 * @see kommunikation
 * @see WbObjekt
 *
 */
public class Weltbild
{
    /** Die Objekte des Weltbilds werden global deklariert, sind also im ganzen
    Weltbild sichtbar, damit jede Methode dieselben <code>WbObjekte</code>
    verwendet und aktualisiert. */
    String teamName;
    WbBallObjekt wbball;
    WbSpielerObjekt wbspieler;
    WbSpielObjekt wbspiel;
    WbAndereSpielerObjekt[] andereSpieler;
    WbRufObjekt wbruf;
    WbObjekt[] weltBild;

    /** fr ungltige Werte von Variablen: */
    final int LEER = -1000;
    final double FEHLT = -1000;

    /** Zahlenwerte fr die Spielseiten: */
    final int left = 1, right = -1;


    /**
     * der Konstruktor erzeugt ein neues Objekt
     * <code>Weltbild</code> mit einem <code>WbBallObjekt</code>,
     * einem <code>WbSpielObjekt</code>,
     * einem <code>WbSpielerObjekt</code> sowie
     * einem Array von <code>WbAndereSpielerObjekten</code>
     * und setzt die Eigenschaften dieser Objekte auf die
     * Ausgangswerte.
     *
     * @param teamName der Name des eigenen Teams
     *
     */
    public Weltbild(String teamName)
    {
      this.teamName = teamName;
      weltBild = new WbObjekt[25];
      weltBild[0] = wbspieler;
      weltBild[1] = wbball;
      weltBild[2] = wbspiel;
      weltBild[24] = wbruf;

      wbspieler = new WbSpielerObjekt(0,LEER,LEER,LEER,LEER,0,0,0,LEER,LEER,0,null,null);
      wbball = new WbBallObjekt(0,57.5,39,0,0,LEER,LEER,LEER);
      wbspiel = new WbSpielObjekt(0,0,"before_kick_off",0,0);
      andereSpieler = new WbAndereSpielerObjekt[21];
      wbruf = null;

      for(int i=0; i<21; i++)
      {
        weltBild[i+3] = andereSpieler[i];
      }
    }


    /**
     * Die Methode aktualisiert die globalen <code>WbObjekte</code>
     * mit der 'init'-Antwort, die vom Server kommt, wenn der
     * Spieler <code>Loki</code> sich dort anmeldet.
     * <br>
     *
     * @param init Diese <code>aktualisiere</code>-Methode wird
     * mit einem <code>initObject</code> aufgerufen.
     *
     * @return Der Rckgabewert der Funktion ist
     * ein Array von <code>WbObjekten</code>. Dieses Array
     * beinhaltet die aktualisierten Werte des Spiels.
     *
     * @exception npe_init kann theoretisch auftreten, wenn das
     * <code>initObject</code> an einer Stelle auf
     * <code>null</code> zeigt.
     *
     */
    public WbObjekt[] aktualisiere(initObject init)
    {
      // alte Werte lschen:
      wbruf = null;
      for(int i=0; i<21; i++)
      {
        andereSpieler[i] = null;
      }

      try
      {
        // aktualisiere zuerst die Spielseite:
        if(init.side().equals("l"))
        {
          wbspiel.seite = left;
        }
        else if(init.side().equals("r"))
        {
          wbspiel.seite = right;
        }
        else
        {
          wbspiel.seite = 0;
          System.out.println("Weltbild: Es wurde keine gltige Spielseite angegeben ! ");
        }

        //dann den Rest:
        wbspieler.nummer = init.uniformNumber();
        wbspiel.status = init.playMode();
        }
        catch(NullPointerException npe_init)
        {
          System.out.println("Leeres initObject im Weltbild angekommen " + npe_init);
        }

        // aktuelle Werte zurckgeben:
        WbObjekt[] aktBild = new WbObjekt[25];
        aktBild[0] = wbspieler;
        aktBild[1] = wbball;
        aktBild[2] = wbspiel;
        for(int i=0; i<21; i++)
        {
          aktBild[i+3] = andereSpieler[i];
        }
        aktBild[24] = wbruf;

        return aktBild;
    }


    /**
     * Mit dieser Methode kann das Weltbild aktualisiert werden.
     * <br>
     * Es wird beim Aufruf ein <code>seeObject</code> bergeben.
     * <br>
     * In dieser Methode werden die </code>seeObjecte</code>
     * je nach Art (Linie, Flag, Spieler, Ball etc.) in eine
     * spezielle <code>ArrayList</code> geschrieben.
     * Ist das fr jedes <code>seeObject</code> geschehen, werden
     * die weiter verarbeitenden Methoden mit diesen
     * <code>ArrayListen</code> als bergabewerte aufgerufen.
     * <br>
     * Daraufhin werden alle <code>WbObjekte</code>, e
     * vtl. aktualisiert, in ein Array gepackt und an die
     * aufrufende Methode zurckgegeben.
     * <br>
     *
     * @param seeObjects Diese <code>aktualisiere</code>-Methode wird
     * mit einem Array von <code>seeObjecten</code> aufgerufen.
     *
     * @return Der Rckgabewert der Funktion ist
     * ein Array von <code>WbObjekten</code>. Dieses Array
     * beinhaltet die aktualisierten Werte des Spiels.
     *
     * @exception npe1 kann auftreten, wenn das Array der
     * <code>seeObjecte</code> leer ist.
     *
     * @exceptionnpe2 kann auftreten, wenn ein
     * <code>seeObject</code> des Arrays leer ist.
     *
     */
    public WbObjekt[] aktualisiere(seeObject[] seeObjects)
    {
      // alte Werte lschen:
      wbruf = null;
      for(int i=0; i<21; i++)
      {
        andereSpieler[i] = null;
      }
      seeObject seeBall = null;

      try
      {
        int seeLaenge = seeObjects.length;
      }
      catch(NullPointerException npe1)
      {
        System.out.println("Weltbild: Fehler: leeres SeeObject : " + npe1);
        return weltBild;
      }

      /* Anlegen der Listen, in die die gesehenen Objekte je nach Typ
      aufgeteilt werden: */
      ArrayList ballList = new ArrayList();
      ArrayList flagList = new ArrayList();
      ArrayList playerList = new ArrayList();
      ArrayList lineList = new ArrayList();
      int ballListIndex = 0;
      int flagListIndex = 0;
      int playerListIndex = 0;
      int lineListIndex = 0;

      /* Jetzt werden die gesehenen Elemente nacheinander angeschaut und je
      nach Typ in die entsprechende Liste geschrieben: */
      for(int i=0; i<seeObjects.length; i++)
      {
        try
        {
          String string1 = seeObjects[i].objName(0);

          if(string1.equals("g") || string1.equals("f"))
          {
            flagList.add(flagListIndex,seeObjects[i]);
            flagListIndex++;
          }
          else if(string1.equals("b") || string1.equals("B"))
          {
            ballList.add(ballListIndex,seeObjects[i]);
            ballListIndex++;
            seeBall = seeObjects[i];
          }
          else if(string1.equals("p") || string1.equals("P"))
          {
            playerList.add(playerListIndex,seeObjects[i]);
            playerListIndex++;
          }
          else if(string1.equals("l"))
          {
            lineList.add(lineListIndex,seeObjects[i]);
            lineListIndex++;
          }
        }
        /* falls ein seeObject leer ist, wird das abgefangen: */
        catch(NullPointerException npe2)
        {
          break;
        }
      }

      /* Jetzt wird fr jede Liste die ensprechende weiterverarbeitende
      Methode aufgerufen: */
      richtungsBestimmung(lineList, lineListIndex);
      positionsBestimmung(flagList, flagListIndex);
      ballWerte(ballList, ballListIndex);
      andereSpielerWerte(playerList, playerListIndex);

      // aktuelle Werte zurckgeben:
      WbObjekt[] aktBild = new WbObjekt[25];
      aktBild[0] = wbspieler;
      aktBild[1] = wbball;
      aktBild[2] = wbspiel;
      for(int i=0; i<21; i++)
      {
        aktBild[i+3] = andereSpieler[i];
      }
      aktBild[24] = wbruf;

      return aktBild;
    }


    /**
     * Mit dieser Funktion kann das Weltbild aktualisiert werden.
     * Sie wird mit einem <code>sense_bodyObject</code> als bergabewert
     * aufgerufen.
     * <br>
     * Dann wird dann die weiter verarbeitende Funktion aufgerufen.
     * Daraufhin werden alle <code>WbObjekte</code> evtl. aktualisiert in
     * ein Array gepackt und an die aufrufende Methode zurckgegeben.
     * <br>
     *
     * @param sensespieler Diese <code>aktualisiere</code>-Methode wird
     * mit einem <code>sense_bodyObject</code> aufgerufen.
     *
     * @return Der Rckgabewert der Funktion ist
     * ein Array von <code>WbObjekten</code>. Dieses Array
     * beinhaltet die aktualisierten Werte des Spiels.
     *
     */
    public WbObjekt[] aktualisiere(sense_bodyObject sensespieler)
    {
      // alte Werte lschen:
      wbruf = null;
      for(int i=0; i<21; i++)
      {
        andereSpieler[i] = null;
      }

      /* zur Weiterverarbeitung vom sense_bodyObject wird die Methode
      <code>spielerWerte</code> verwendet: */
      spielerWerte(sensespieler);

      // aktuelle Werte zurckgeben:
      WbObjekt[] aktBild = new WbObjekt[25];
      aktBild[0] = wbspieler;
      aktBild[1] = wbball;
      aktBild[2] = wbspiel;
      for(int i=0; i<21; i++)
      {
        aktBild[i+3] = andereSpieler[i];
      }
      aktBild[24] = wbruf;

      return aktBild;
    }


    /**
     * Mit dieser Funktion kann das Weltbild aktualisiert werden.
     * Es wird ein <code>hearObject</code> an diese Funktion geschickt.
     * In dieser wird dann die weiter verarbeitende Funktion aufgerufen.
     * <br>
     * Daraufhin werden alle <code>WbObjekte</code>, evtl. aktualisiert, in
     * ein Array gepackt und an die aufrufende Methode zurckgegeben.
     * <br>
     *
     * @param hearspieler Diese <code>aktualisiere</code>-Methode wird
     * mit einem <code>hearObject</code> aufgerufen.
     *
     * @return Der Rckgabewert der Funktion ist
     * ein Array von <code>WbObjekten</code>. Dieses Array
     * beinhaltet die aktualisierten Werte des Spiels.
     *
     * @exception npe_hear_sender Der Fall eines leeren
     * <code>hearObjects</code> wird abgefangen.
     *
     */
    public WbObjekt[] aktualisiere(hearObject hearspieler)
    {
      // alte Werte lschen:
      wbruf = null;
      for(int i=0; i<21; i++)
      {
        andereSpieler[i] = null;
      }

      /* Hier wirde unterschieden, ob der Ruf vom Schiedsrichter kommt oder
      von einem andern Spieler und die weiter verarbeitende Methode
      aufgerufen. */
      try
      {
        if(hearspieler.sender().equals("Direction"))
        {
          spielerRuf(hearspieler);
        }
        else if( hearspieler.sender().equals( "referee" ) )
        {
          wbspiel.status = hearspieler.message();
          if( hearspieler.message().equals("half_time") || hearspieler.message().equals("time_extended") )
          {
            // es gibt scheinbar keinen Seitenwechsel. sonst: wbspiel.seite = -(wbspiel.seite);
          }
        }
      }
      catch(NullPointerException npe_hear_sender)
      {
        // Der Fall eines leeren hearObjects wird abgefangen.
      }

      // aktuelle Werte zurckgeben:
      WbObjekt[] aktBild = new WbObjekt[25];
      aktBild[0] = wbspieler;
      aktBild[1] = wbball;
      aktBild[2] = wbspiel;
      for(int i=0; i<21; i++)
      {
        aktBild[i+3] = andereSpieler[i];
      }
      aktBild[24] = wbruf;

      return aktBild;
    }


    /**
     * Die Methode <code>richtungsBestimmung</code> wird von der Methode
     * <code>aktualisiere(seeObject[])</code> aufgerufen.
     * <br>
     * Sie verarbeitet die Liste, in der die Linien gespeichert wurden,
     * die der Spieler in diesem Zyklus gesehen hat und sie bestimmt wie der
     * Name schon sagt die (Aus-)Richtung des Spielers in Abhngigkeit von
     * der Spielseite.
     * <br>
     *
     * @param linien Das ist eine der in der
     * <code>aktualisiere(seeObject[])</code>-Methode erzeugten
     * <code>ArrayListen</code>.
     *
     * @param laenge Diese Variable enthlt die Lnge der bergebenen
     * <code>ArrayList</code>.
     *
     * @exception ioobe_linien fngt den Fall ab, da der Spieler in diesem
     * Zyklus keine Linie(n) sieht.
     *
     */
    public void richtungsBestimmung(ArrayList linien, int laenge)
    {
      try
      {
        double direction = ((seeObject)linien.get(0)).direction();
        if( ((seeObject)linien.get(0)).objName(1).equals("t") )
        {
          if( direction >= 0 )
          {
            wbspieler.richtung = bereinigt( -direction - wbspieler.relative_blickrichtung() );
          }
          else
          {
            wbspieler.richtung = bereinigt( -180 - direction - wbspieler.relative_blickrichtung() );
          }
        }
        else if( ((seeObject)linien.get(0)).objName(1).equals("b") )
        {
          if( direction >= 0 )
          {
            wbspieler.richtung = bereinigt( 180 - direction - wbspieler.relative_blickrichtung() );
          }
          else
          {
            wbspieler.richtung = bereinigt( -direction - wbspieler.relative_blickrichtung() );
          }
        }
        else if( ((seeObject)linien.get(0)).objName(1).equals("r") )
        {
          if(direction >= 0)
          {
            wbspieler.richtung = bereinigt(90 - direction - wbspieler.relative_blickrichtung() );
          }
          else
          {
            wbspieler.richtung = bereinigt(-90 - direction - wbspieler.relative_blickrichtung() );
          }
        }
        else //( ((seeObject)linien.get(0)).objName(1).equals("l") )
        {
          if(direction >= 0)
          {
            wbspieler.richtung = bereinigt(-90 - direction - wbspieler.relative_blickrichtung() );
          }
          else
          {
            wbspieler.richtung = bereinigt(90 - direction - wbspieler.relative_blickrichtung() );
          }
        }
        if(wbspiel.seite() == -1)
        {
          wbspieler.richtung = bereinigt( 180 + wbspieler.richtung );
        }
        wbspieler.blickrichtung = bereinigt( wbspieler.richtung + wbspieler.relative_blickrichtung );
      }
      catch(IndexOutOfBoundsException ioobe_linien)
      {
        wbspieler.richtung = -1000;
      }

    }


    /**
     * Hier passiert nicht viel.
     * Der Ruf eines anderen Spielers wird gespeichert.
     * <br>
     *
     * @param ruf Das ist der Ruf eines Spielers, also kein Schiedsrichterruf.
     *
     */
    public void spielerRuf(hearObject ruf)
    {
      wbruf = new WbRufObjekt(ruf.time(), ruf.vonUns(), ruf.direction(), ruf.message());
    }


  /**
   * Ein <code>sense_bodyObjekt</code> wird in dieser Methode weiter
   * verarbeitet. Hier wird das <code>WbSpielerObjekt</code>
   * des <code>Weltbilds</code> auf den aktuellen Stand gebracht.
   * <br>
   *
   * @param sensespieler beinhaltet die Werte des Spielers selbst, wie
   * Blickrichtung, Energie etc.
   *
   * @exception npe_sense Der theoretische Fall eines leeren <code>sense_bodyObjects</code>
   * wird abgefangen.
   *
   */
  public void spielerWerte(sense_bodyObject sensespieler)
  {
    try
    {
      wbspieler.zyklus = sensespieler.time();
      wbspieler.geschwindigkeitX = Math.cos(wbspieler.richtung) * sensespieler.amountOfSpeed();
      wbspieler.geschwindigkeitY = Math.sin(wbspieler.richtung) * sensespieler.amountOfSpeed();
      wbspieler.relative_blickrichtung = sensespieler.headDirection();
      wbspieler.blickrichtung = bereinigt( wbspieler.richtung() + wbspieler.relative_blickrichtung() );
      wbspieler.recovery = sensespieler.stamina();
      wbspieler.effort = sensespieler.effort();
      wbspieler.viewQuality = sensespieler.viewQuality();
      wbspieler.viewWidth = sensespieler.viewMode();
    }
    catch(NullPointerException npe_sense)
    {
      // Der Fall eines leeren sense_bodyObjects wird abgefangen.
      System.out.println("Weltbild: Fehler npe_sense: " + npe_sense);
    }
  }


  /**
   * Diese Methode wird von der Methode
   * <code>aktualisiere(seeObject[])</code> aufgerufen.
   * <br>
   * Sie verarbeitet die Liste, in der die Spieler gespeichert wurden,
   * die der Spieler in diesem Zyklus gesehen hat und sie bestimmt alle
   * ntigen Eigenschaften der Spieler, wie Teamzugehrigkeit,
   * Entfernung etc.
   * <br>
   *
   * @param linien Das ist eine der in der
   * <code>aktualisiere(seeObject[])</code>-Methode erzeugten
   * <code>ArrayListen</code>.
   *
   * @param laenge Diese Variable enthlt die Lnge der bergebenen
   * <code>ArrayList</code>.
   *
   * @exception e_andereSpielerWerte fngt einen Fehler bei der Bestimmung
   * der Eigenschaften der gesehenen Spieler ab.
   *
   */
  public void andereSpielerWerte(ArrayList spieler, int laenge)
  {
    for(int i=0; i<laenge; i++)
    {
      try
      {
        /* Zunchst werden fr jeden Spieler die Eigenschaftswerte auf
        Defaultwerte gesetzt. */

        int nummer = -1;
        int mannschaft = 0;
        int zyklus = ((seeObject)spieler.get( i )).time();

	double distance = ((seeObject)spieler.get(i)).distance();
	double direction = ((seeObject)spieler.get(i)).direction();

	double abstandX, abstandY;
	double winkel_ = Math.toRadians( bereinigt( wbspieler.blickrichtung() + direction ) );
	double betrag_winkel_ = Math.abs( winkel_ );

        // bestimme den X-Abstand zum Objekt:
        abstandX = wbspiel.seite() * ( Math.cos( winkel_ ) * distance );
        // bestimme den Y-Abstand zum Objekt:

        if(wbspiel.seite()*winkel_ >= 0)
        {
          abstandY = Math.sin( betrag_winkel_ ) * distance;
        }
        else
        {
          abstandY = -Math.sin( betrag_winkel_ ) * distance;
        }

        // damit wir "schoenere" werte bekommen...
        if( Math.abs( Math.rint( abstandX ) - abstandX ) < 0.0000001 ) { abstandX = Math.rint( abstandX ); }
        if( Math.abs( Math.rint( abstandY ) - abstandY ) < 0.0000001 ) { abstandY = Math.rint( abstandY ); }

        double positionX = wbspieler.positionX() + abstandX;
        double positionY = wbspieler.positionY() + abstandY;
        // ! david - es wird noch keine Geschwindigkeit berechnet
        double geschwindigkeitX = FEHLT;
        double geschwindigkeitY = FEHLT;
        boolean torwart = false;
        double ballEntfernung = FEHLT;
        double entfernung = Math.sqrt( Math.pow(wbspieler.positionX - positionX, 2) + Math.pow(wbspieler.positionY - positionY, 2) );

        // Jetzt wird der Spieler mit diesen Defaultwerten erzeugt:
        andereSpieler[i] = new WbAndereSpielerObjekt(zyklus, nummer, mannschaft, positionX, positionY, geschwindigkeitX, geschwindigkeitY, torwart, entfernung, ballEntfernung);

        /* Zuletzt werden die tatschlichen Eigenschaftswerte abgefragt und
        gespeichert. Dabei wird abgebrochen, wenn eine Eigenschaft 'null' ist
        oder wenn alle Eigenschaften abgearbeitet sind, wird mit dem nchsten
        Spieler fortgefahren. Durch die vorherige initialisierung mit
        Defaultwerten stimmen die Eigenschaften dann auch bei einem
        eventuellen Abbruch. */
        if(wbball.positionX >= 0)
        {
          andereSpieler[i].ballEntfernung = Math.sqrt( Math.pow(wbball.positionX - andereSpieler[i].positionX, 2) + Math.pow(wbball.positionY - andereSpieler[i].positionY, 2) );
        }

        String mannschaftsName = ((seeObject)spieler.get(i)).objName(1);

        if( mannschaftsName.equals( teamName ) )
        {
          andereSpieler[i].mannschaft = 1; // eigene Mannschaft
        }
        else if( !mannschaftsName.equals( "OUT_OF_BOUNDS" ) )
        {
          andereSpieler[i].mannschaft = -1; // gegnerische Mannschaft
        }
        else
        {
          andereSpieler[i].mannschaft = 0; // gegnerische Mannschaft
        }

        andereSpieler[i].nummer = Integer.parseInt( ((seeObject)spieler.get(i)).objName(2) );
        String torwart_ = ((seeObject)spieler.get(i)).objName(3);
        if(torwart_.equals("goalie"))
        {
          andereSpieler[i].torwart = true;
        }
      }
      catch(Exception e_andereSpielerWerte)
      {

      }

    }
  }


  /**
   *
   * Diese Methode wird von der Methode
   * <code>aktualisiere(seeObject[])</code> aufgerufen.
   * <br>
   * Sie verarbeitet die Liste, in welcher der Ball gespeichert wurdeund sie
   * bestimmt alle bentigten Eigenschaften des Balls, wie Entfernung,
   * Geschwindigkeit etc.
   * <br>
   *
   * @param visualball Das ist eine der in der
   * <code>aktualisiere(seeObject[])</code>-Methode erzeugten
   * <code>ArrayListen</code>.
   *
   * @param laenge Diese Variable enthlt die Lnge der bergebenen
   * <code>ArrayList</code>.
   *
   */
  public void ballWerte(ArrayList visualball, int laenge)
  {
    if( laenge == 1 )
    {
      double altePositionX = wbball.positionX();
      double altePositionY = wbball.positionY();
      double distance = ((seeObject)visualball.get(0)).distance();
      double direction = ((seeObject)visualball.get(0)).direction();

      double abstandX, abstandY;
      wbball.winkel = bereinigt( wbball.relativer_winkel + wbspieler.blickrichtung );
      double winkel_ = Math.toRadians( wbball.winkel );
      double betrag_winkel_ = Math.abs( winkel_ );

      // bestimme den X-Abstand zum Objekt:
      abstandX = wbspiel.seite() * ( Math.cos( winkel_ ) * distance );
      // bestimme den Y-Abstand zum Objekt:

      if(wbspiel.seite()*winkel_ >= 0)
      {
        abstandY = Math.sin( betrag_winkel_ ) * distance;
      }
      else
      {
        abstandY = -Math.sin( betrag_winkel_ ) * distance;
      }

      // damit wir "schoenere" werte bekommen...
      if( Math.abs( Math.rint( abstandX ) - abstandX ) < 0.0000001 ) { abstandX = Math.rint( abstandX ); }
      if( Math.abs( Math.rint( abstandY ) - abstandY ) < 0.0000001 ) { abstandY = Math.rint( abstandY ); }

      wbball.positionX = wbspieler.positionX() + abstandX;
      wbball.positionY = wbspieler.positionY() + abstandY;
      wbball.geschwindigkeitX = ( wbball.positionX() - altePositionX );
      wbball.geschwindigkeitY = ( wbball.positionY() - altePositionY );


      wbball.relativer_winkel = ((seeObject)visualball.get(0)).direction();
      wbball.entfernung = ((seeObject)visualball.get(0)).distance();
    }
    else
    {
      wbball.positionX = LEER;
      wbball.positionY = LEER;
    }
  }


  /**
   * Diese Methode wird von der Methode
   * <code>aktualisiere(seeObject[])</code> aufgerufen.
   * <br>
   * Sie errechnet aus mehreren <code>seeObjecten</code> die aktuelle
   * Position des Spielers, genauer gesagt mit den Fahnen und ggf. einem Tor,
   * die der Spieler in diesem Zyklus gesehen hat und die zuvor fr diese
   * Verarbeitung in dieser Funktion (getrennt von den anderen Objekten) in
   * eine separate Liste 'flagList' geschrieben wurden.
   * <br>
   * Dazu greift diese Funktion auf die <code>koordinaten</code> dieser
   * Objekte zu, die von der gleichnamigen Methode geliefert werden.
   * <br>
   *
   * @param fahnen Das ist eine der in der
   * <code>aktualisiere(seeObject[])</code>-Methode erzeugten
   * <code>ArrayListen</code>.
   *
   * @param laenge Diese Variable enthlt die Lnge der bergebenen
   * <code>ArrayList</code>.
   *
   */
  public void positionsBestimmung(ArrayList fahnen, int laenge)
  {
    double[][] a = new double[laenge][];
    double[] deltaX = new double[laenge];
    double[] deltaY = new double[laenge];
    double[] X = new double[laenge];
    double[] Y = new double[laenge];
    double[] distance = new double[laenge];

    for(int i=0; i<laenge; i++)
    {
      a[i] = koordinaten( ((seeObject)fahnen.get(i)).objName() );
      distance[i] = ((seeObject)fahnen.get(i)).distance();
      double direction = ((seeObject)fahnen.get(i)).direction();

      double winkel_ = Math.toRadians( bereinigt( wbspieler.blickrichtung() + direction ) );
      double betrag_winkel_ = Math.abs( winkel_ );

      // bestimme den X-Abstand zum Objekt:
      deltaX[i] = wbspiel.seite() * ( Math.cos( winkel_ ) * distance[i] );

      // bestimme den Y-Abstand zum Objekt:
      if(wbspiel.seite()*winkel_ >= 0)
      {
        deltaY[i] = Math.sin( betrag_winkel_ ) * distance[i];
      }
      else
      {
        deltaY[i] = -Math.sin( betrag_winkel_ ) * distance[i];
      }

      // damit wir 'schnere' werte bekommen...
      if( Math.abs( Math.rint( deltaX[ i ] ) - deltaX[ i ] ) < 0.0000001 ) { deltaX[ i ] = Math.rint( deltaX[ i ] ); }
      if( Math.abs( Math.rint( deltaY[ i ] ) - deltaY[ i ] ) < 0.0000001 ) { deltaY[ i ] = Math.rint( deltaY[ i ] ); }

      X[i] = a[i][0] - deltaX[i];
      Y[i] = a[i][1] - deltaY[i];

    }

    // errechne den Mittelwert:
    double gewichteteSummeX = 0, gewichteteSummeY = 0, summeGewichte = 0;

    for(int i=0; i<laenge; i++)
    {
      if( (X[i] >= 0) && (Y[i] >= 0) )
      {
        gewichteteSummeX = gewichteteSummeX + ( X[i] / distance[i] );
        gewichteteSummeY = gewichteteSummeY + ( Y[i] / distance[i] );
        summeGewichte = summeGewichte + ( 1 / distance[i] );
      }
    }

    wbspieler.positionX = gewichteteSummeX / summeGewichte;
    wbspieler.positionY = gewichteteSummeY / summeGewichte;
  }


  /**
   * Die Methode <code>koordinaten</code> gibt die absoluten
   * X- und Y-Koordinaten fr ein bestimmtes flag oder goal zurck.
   * <br>
   *
   * @param seen Es wird ein bestimmtes Objekt bergeben, von dem die
   * Koordinaten bestimmt werden sollen.
   *
   * @return koordinaten gibt die X- und die Y-Position des angepeilten
   * Objekts zurck.
   *
   */
  public double[] koordinaten(String[] seen)
  {
    double[] koordinaten = new double[2];
    double x = -1, y = -1;

    // goal Objekt:
    if(seen[0].equals("g"))
    {
      y = 39;
      if(seen[1].equals("l")) // goal l
      {
        x = 5;
      }
      else if(seen[1].equals("r")) // goal r
      {
        x = 110;
      }
      else
      {
        x = -1;
      }
    }

    // flag Objekt:
    else if(seen[0].equals("f"))
    {
      if(seen[1].equals("c")) // flag c
      {
        x = 57.5;
        if(seen.length == 2)
        {
          y = 39;
        }
        else if(seen.length == 3)
        {
          if(seen[2].equals("t")) // flag c t
          {
            y = 5;
          }
          else if(seen[2].equals("b")) // flag c b
          {
            y = 73;
          }
          else // flag c -
          {
            y = -1;
          }
        }
        else
        {
          y = -1;
        }
      }

      else if(seen[1].equals("t")) // flag t
      {
        y = 0;
        if(seen.length == 3)
        {
          x = 57.5;
        }
        else if(seen.length == 4)
        {
          if(seen[2].equals("l")) //flag t l
          {
            if(seen[3].equals("10"))
            {
              x = 57.5 - 10;
            }
            else if(seen[3].equals("20"))
            {
              x = 57.5 - 20;
            }
            else if(seen[3].equals("30"))
            {
              x = 57.5 - 30;
            }
            else if(seen[3].equals("40"))
            {
              x = 57.5 - 40;
            }
            else if (seen[3].equals("50"))
            {
              x = 57.5 - 50;
            }
            else
            {
              x = -1;
            }
          }
          else if(seen[2].equals("r")) //flag t r
          {
            if(seen[3].equals("10"))
            {
              x = 57.5 + 10;
            }
            else if(seen[3].equals("20"))
            {
              x = 57.5 + 20;
            }
            else if(seen[3].equals("30"))
            {
              x = 57.5 + 30;
            }
            else if(seen[3].equals("40"))
            {
              x = 57.5 + 40;
            }
            else if(seen[3].equals("50"))
            {
              x = 57.5 + 50;
            }
            else
            {
              x = -1;
            }
          }
          else
          {
            x = -1;
          }
        }
        else
        {
          x = -1;
        }
      }

      else if(seen[1].equals("b")) // flag b
      {
        y = 78;
        if(seen.length == 3)
        {
          x = 57.5;
        }
        else if(seen.length == 4)
        {
          if(seen[2].equals("l")) //flag b l
          {
            if(seen[3].equals("10"))
            {
              x = 57.5 - 10;
            }
            else if(seen[3].equals("20"))
            {
              x = 57.5 - 20;
            }
            else if(seen[3].equals("30"))
            {
              x = 57.5 - 30;
            }
            else if(seen[3].equals("40"))
            {
              x = 57.5 - 40;
            }
            else if(seen[3].equals("50"))
            {
              x = 57.5 - 50;
            }
            else
            {
              x = -1;
            }
          }
          else if(seen[2].equals("r")) //flag b r
          {
            if(seen[3].equals("10"))
            {
              x = 57.5 + 10;
            }
            else if(seen[3].equals("20"))
            {
              x = 57.5 + 20;
            }
            else if(seen[3].equals("30"))
            {
              x = 57.5 + 30;
            }
            else if(seen[3].equals("40"))
            {
              x = 57.5 + 40;
            }
            else if(seen[3].equals("50"))
            {
              x = 57.5 + 50;
            }
            else
            {
              x = -1;
            }
          }
          else
          {
            x = -1;
          }
        }
        else
        {
          x = -1;
        }
      }

      else if(seen[1].equals("p")) // flag p
      {
        if(seen[2].equals("l")) // flag p l
        {
          x = 16+5; //stimmt das ???
          if(seen[3].equals("t"))
          {
            y = 19;
          }
          else if(seen[3].equals("c"))
          {
            y = 39;
          }
          else if(seen[3].equals("b"))
          {
            y = 59;
          }
          else
          {
            y = -1;
          }
        }
        else if(seen[2].equals("r")) // flag p r
        {
          x = 110-16; // stimmt das ???
          if(seen[3].equals("t"))
          {
            y = 19;
          }
          else if(seen[3].equals("c"))
          {
            y = 39;
          }
          else if(seen[3].equals("b"))
          {
            y = 59;
          }
          else
          {
            y = -1;
          }
        }
        else
        {
          x = -1;
          y = -1;
        }
      }

      else if(seen[1].equals("r")) // flag r
      {
        if(seen.length == 3)
        {
          if(seen[2].equals("t")) // flag r t
          {
            x = 110;
            y = 5;
          }
          else if(seen[2].equals("b"))
          {
            x = 110;
            y = 73;
          }
          else if(seen[2].equals("0"))
          {
            x = 115;
            y = 39;
          }
          else
          {
            x = -1;
            y = -1;
          }
        }
        else if(seen.length == 4)
        {
          x = 115;
          if(seen[2].equals("t")) // flag r t *
          {
            if(seen[3].equals("10"))
            {
              y = 39 - 10;
            }
            else if(seen[3].equals("20"))
            {
              y = 39 - 20;
            }
            else if(seen[3].equals("30"))
            {
              y = 39 - 30;
            }
            else
            {
              y = -1;
            }
          }
          else if(seen[2].equals("b")) // flag r b *
          {
            if(seen[3].equals("10"))
            {
              y = 39 + 10;
            }
            else if(seen[3].equals("20"))
            {
              y = 39 + 20;
            }
            else if(seen[3].equals("30"))
            {
              y = 39 + 30;
            }
            else
            {
              y = -1;
            }
          }
          else
          {
            y = -1;
          }
        }
        else
        {
          x = -1;
          y = -1;
        }
      }

      else if(seen[1].equals("l"))
      {
        if(seen.length == 3)
        {
          if(seen[2].equals("t")) // flag l t
          {
            x = 5;
            y = 5;
          }
          else if(seen[2].equals("b"))
          {
            x = 5;
            y = 73;
          }
          else if(seen[2].equals("0"))
          {
            x = 0;
            y = 39;
          }
          else
          {
            x = -1;
            y = -1;
          }
        }
        else if(seen.length == 4)
        {
          x = 0;
          if(seen[2].equals("t")) // flag l t *
          {
            if(seen[3].equals("10"))
            {
              y = 39 - 10;
            }
            else if(seen[3].equals("20"))
            {
              y = 39 - 20;
            }
            else if(seen[3].equals("30"))
            {
              y = 39 - 30;
            }
            else
            {
              y = -1;
            }
          }
          else if(seen[2].equals("b")) // flag l b *
          {
            if(seen[3].equals("10"))
            {
              y = 39 + 10;
            }
            else if(seen[3].equals("20"))
            {
              y = 39 + 20;
            }
            else if(seen[3].equals("30"))
            {
              y = 39 + 30;
            }
            else
            {
              y = -1;
            }
          }
          else
          {
            y = -1;
          }
        }
        else
        {
          x = -1;
          y = -1;
        }
      }

      else if(seen[1].equals("g"))
      {
        if(seen[2].equals("r")) // flag g r *
        {
          x = 110;
          if(seen[3].equals("t"))
          {
            y = 39 - 7.01;
          }
          else if(seen[3].equals("b"))
          {
            y = 39 + 7.01;
          }
          else
          {
            y = -1;
          }
        }
        else if(seen[2].equals("l")) // flag g l *
        {
          x = 5;
          if(seen[3].equals("t"))
          {
            y = 39 - 7.01;
          }
          else if(seen[3].equals("b"))
          {
            y = 39 + 7.01;
          }
          else
          {
            y = -1;
          }
        }
        else
        {
          x = -1;
          y = -1;
        }
      }
      else
      {
        x = -1;
        y = -1;
      }
    }

    else if(seen[0].equals("G")) // Goal
    {
      y = 39;
      x = -1;
    }

    else if(seen[0].equals("F"))
    {
      x = -1;
      y = -1;
    }

    else
    {
      x = 0;
      y = 0;
    }

    koordinaten[0] = x;
    koordinaten[1] = y;
    return koordinaten;
  }

  /**
   * Diese Funktion rechnet den Winkel um, falls er betragsmig
   * grer als 180 Grad ist.
   * <br>
   *
   * @param wert der umzurechnende Winkel
   *
   * @return wert der umgerechnete Winkel
   *
   */
  private double bereinigt(double wert)
  {
    if(wert > 180)
    {
      wert = wert - 360;
    }
    else if(wert < -180)
    {
      wert = wert + 360;
    }
    return wert;
  }
}

