// kommunikation.java stand: 200200707

/** wird fuer StringTokenizer benoetigt */
import java.util.*;

/**
 * kommunikation: wird von Loki instanziert
 *
 * uebernimmt:
 * <ul>
 * <li>das empfangen der daten vom server
 * <li>die decodierung der events in die entsprechenden objekte
 * <li>das senden der neuen befehle an den server
 * </ul>
 *
 * @author <a href="mailto:rmathes@uni-mannheim.de">Rene M</a>
 * @version 0.0.1
 * @see <code>Loki</code>
 * @see <code>Weltbild</code>
 * @see <code>Strategie</code>
 *
 * <br>-----<br>
 * <br>
 * die kommunikation wird instanziert mit kommunikation kom = new kommunikation( serverName, port, teamName );<br>
 * <br>
 * public methoden / variablen:<br>
 * kommunikation( String serverName, int port, String teamName ) - konstruktor<br>
 * bye() - meldet den spieler vom server ab<br>
 * dash( double power ) - laesst den spieler rennen<br>
 * fange( dounle direction ) - der torhueter versucht den ball zu fangen<br>
 * getXObject() - liefert ein erzeugtes xObject zurueck<br>
 * init() - meldet den spieler am server an<br>
 * init( boolean goalie ) - meldet den spieler als goalie am server an<br>
 * kick( double power, double direction ) - tritt den ball<br>
 * move( double x, double y ) - bewegt den spieler auf eine feste koordinate<br>
 * neuesXObject() - ist ein neues xObject vorhanden?<br>
 * turn( double moment ) - der spieler dreht sich<br>
 *<br>
 *<br>-----<br>
 *<br>
 * private methoden / variablen:<br>
 * bearbeiteServerDaten( String serverDaten ) - parst die empfangenen daten in
 *                                              die einzelnen objekte<br>
 * CommunicationChannel cc - udp traffic mit dem server<br>
 * String serverName - fuer communication channel<br>
 * int port - fuer communication channel<br>
 * String teamName - um das team anzumelden<br>
 *<br>
 * um nur mal einige aufzuzaehlen ;)<br>
*/
public class kommunikation {

// globale private variablen
/** zum Abfragen der verschiedenen objekte */
private boolean neuesHearObject = false;
private boolean neuesSeeObject = false;
private boolean neuesSense_bodyObject = false;
private boolean neuesInitObject = false;
private boolean neuesErrorObject = false;

/** zum zwischenspeichern der verschiedenen objekte */
private hearObject ho;
private seeObject[] so;
private sense_bodyObject s_bo;
private initObject io;
private errorObject eo;

/** der CommunicationChannel */
private CommunicationChannel cc;
/** name oder ip des servers */
private String serverName;
/** port des servers an den connected werden soll */
private int port;
// ^^ wird verwendet um CommunicationChannel zu instanzieren.
/** name des teams */
private String teamName;
// ^^ wird verwendet um den spieler dem team zuzuordnen.

private static final double FEHLT = -1000;

// ----------

/**
 * instanziert die <code>kommunikation</code> und setzt variablen, der <code>CommunicationChannel</code> wird erst im konstruktor erzeugt!
 *
 * @param serverName name oder ip des servers
 * @param port port an den connected werden soll
 * @param teamName name des teams
 */
public kommunikation( String serverName, int port, String teamName ) {
  this.serverName = serverName;
  this.port = port;
  this.teamName = teamName;
  System.out.println( "Kommunikation: Verbindung wird aufgebaut zu: rcss://"+serverName+":"+port+"/"+teamName );

/** <code>CommunicationChannel</code> erzeugen */
  cc = new CommunicationChannel( serverName, port );
// cc instanziert
}
// ende konstruktor


// -----

/**
 * getObjects() - empfaengt die daten vom communicationchannel und parst sie
 * in die entsprechenden objekte...
 */
public void getObjects() {
  bearbeiteServerDaten( cc.receive() );
}

// -----

/**
 * Methoden, die die verschiedenen Objekte zurueckliefern
 */

/**
 * getHearObject() - liefert das hearObject zurueck
 * @return <code>hearObject</code>
 */
public hearObject getHearObject() {
  neuesHearObject = false;
  return ho;
}

/**
 * getSeeObject() - liefert ein array seeObject[] zurueck
 * @return <code>seeObject[]</code>
 */
public seeObject[] getSeeObject() {
  neuesSeeObject = false;
  return so;
}

/**
 * getSense_bodyObject() - liefert das sense_bodyObject zurueck
 * @return <code>sense_bodyObject</code>
 */
public sense_bodyObject getSense_bodyObject() {
  neuesSense_bodyObject = false;
  return s_bo;
}

/**
 * getInitObject() - liefert das initObject zurueck
 * @return <code>initObject</code>
 */

public initObject getInitObject() {
  neuesInitObject = false;
  return io;
}

/**
 * getErrorObject() - liefert das errorObject zurueck
 * @return <code>errorObject</code>
 */

public errorObject getErrorObject() {
  neuesErrorObject = false;
  return eo;
}

// -----

/**
 * methoden die die verschiedenen stati zurueckliefern
 */

/**
 * neuesHearObect() - liefert den status der hearObjecte
 */
public boolean neuesHearObject() { return neuesHearObject; }

/**
 * neuesSeeObject() - liefert den status der seeObjecte
 */
public boolean neuesSeeObject() { return neuesSeeObject; }

/**
 * neuesSense_bodyObject() - liefert den status der sense_bodyObjecte
 */
public boolean neuesSense_bodyObject() { return neuesSense_bodyObject; }

/**
 * neuesInitObject() - liefert den status der initObjecte
 */
public boolean neuesInitObject() { return neuesInitObject; }

/**
 * neuesErrorObject() - liefert den status der errorObjecte
 */
public boolean neuesErrorObject() { return neuesErrorObject; }

// -----

/**
 * methoden, die befehle an den server senden
 */

 /**
 * init() - meldet einen spieler am server an.
 * ohne parameter wird "goalie" auf false gesetzt.
 */
public void init() {
  this.init( false );
}
// ende von init()

/**
 * init( boolean goalie ) - meldet den spieler am server an.
 *
 * @param <code>boolean</code>goalie : spieler oder torhueter?
 */
public void init( boolean goalie ) {
  System.out.println( "Kommunikation: melde den Spieler "+( goalie ? "als Goalie ": "" ) +"am Server an..." );
  cc.send( new String( "(init " + teamName + " (version 7.0)" + ( goalie ? " (goalie) " : "" )+ ")" ) );
}
// ende von init( boolean goalie );

/**
 * bye() - meldet den spieler vom server ab.
 */
public void bye() {
  cc.send( new String( "(bye)" ) );
}
// ende von bye()

 /**
 * dash( double power ) - rennt mit der angegebenen kraft in die aktuelle richtung
 *
 * @param power : krafteinsatz des spielers
 */
public void dash( double power ) {
  cc.send( new String( "(dash " + power + ")" ) );
}
// ende von dash( double power )

/**
 * fange( double direction ) - der Torhueter versucht den Ball zu fangen.
 *
 * @param direction : Richtung des "catch" versuches
 */
public void fange( double direction ) {
  cc.send( new String( "(catch " + direction + ")" ) );
}
// ende von fange( double direction )

/**
 * kick( double power, double direction ) - spielt den ball mit gegebener kraft und in die gegebene richtung
 *
 * @param power : krafteinsatz des spielers
 * @param direction : die richtung des kicks
 */
public void kick( double power, double direction ) {
  cc.send( new String( "(kick " + power + " " + direction + ")" ) );
}
// ende von kick( double power, double direction );

/**
 * move( double x, double y ) - setzt den spieler auf eine feste position im spielfeld
 * move ist nicht immer moeglich, sondern nur <ul><li>beim kick-off<li>nach einem tor</ul>
 *
 * @param x : x-position
 * die x position kann im bereich von -52.5 bis +52.5 [in metern] liegen.
 * negativ: eigene haelfte des spielfelds
 * positiv: gegnerische haelfte
 * gemessen wird von der mittellinie aus
 *
 * @param y : y-position
 * die y position kann im bereich von -34.0 bis +34.0 [in metern] liegen.
 * negativ: linke seite des feldes, in blickrichtung auf das gegnerische tor
 * positiv: rechte seite.
 * gemessen wird der abstand von der verbindung beider tore und dem mittelpunkt des feldes
 */
public void move( double x, double y ) {
  x = x - 57.5;
  y = y - 39;
  cc.send( new String( "(move " + x + " " + y + ")" ) );
}
// ende von move( double x, double y )

/**
 * turn( double moment ) - dreht den koerper des spielers um einen bestimmten winkel
 *
 * @param moment : winkel um den gedreht werden soll
 * es wird im uhrzeigersinn gedreht, also winkel > 0 entsprechen einer drehung nach rechts.
 * der absolutbetrag des winkels muss kleiner gleich 180 grad sein, alles groessere wird vom
 * server als 180 grad drehung gewertet.
 */
public void turn( double moment ) {
  cc.send( new String( "(turn " + moment + ")" ) );
}
// ende von turn( double moment )

// ----------

/**
 * private methoden
 */

/**
 * bearbeiteServerDaten( String serverDaten )
 * decodiert den string und erzeugt die Objekte
 *
 * @param serverDaten enthaelt den vom server gelieferten string
 * @version 0.0.1
*/
private void bearbeiteServerDaten( String serverDaten ) {
//  System.out.print( "Kommunikation: Bearbeite empfangene Daten..." );
/** serverDaten decodieren */
  int time = (int)Math.floor( FEHLT );
  String type = "";
/**
 * unterscheidung zweier serverDaten:
 * 1) ohne weitere klammern (ohne zeitangabe)
 * 2) mit weiteren klammern (also auch mit zeitangabe etc)
 */

  if( ( serverDaten.startsWith( "(init" ) ) || ( serverDaten.startsWith( "(error" ) ) || ( serverDaten.startsWith( "(server_param" ) ) || ( serverDaten.startsWith( "(player_param" ) ) || ( serverDaten.startsWith( "(hear" ) ) ) {
// alle typen die keine weiteren klammern und bis auf "hear" keine zeitangabe besitzen...
    serverDaten = serverDaten.substring(1, serverDaten.length() -2 );
    StringTokenizer initErrorString = new StringTokenizer( serverDaten, " " );
    type = initErrorString.nextToken();
/** type = {init|error|hear} etc */

/**  "init" message? */
    if( ( type.startsWith( "init" ) ) && !( neuesInitObject ) ) {
//      System.out.println( "Kommunikation: init message empfangen..." );
// init message bearbeiten
      String side = initErrorString.nextToken().trim();
      int uniformNumber = Integer.parseInt( initErrorString.nextToken().trim() );
      String playMode = initErrorString.nextToken().trim();
      initObject io = new initObject( side, uniformNumber, playMode );
      this.io = io;
      neuesInitObject = true;
    }

/** "error" message? */
    if( ( type.startsWith( "error" ) ) && !( neuesErrorObject ) ) {
//      System.out.println( "Kommunikation: error message empfangen..." );
// error message bearbeiten
      errorObject eo = new errorObject( initErrorString.nextToken() );
      this.eo = eo;
      neuesErrorObject = true;
    }
/** "hear message? */
    if( ( type.startsWith( "hear" ) ) && !( neuesHearObject ) ) {
/**
 * in initErrorString steht jetzt alles nach dem "hear <timestamp>"
 * die gesendete nachricht ohne den typ und den timestamp in die einzelnen parameter aufteilen
 * und ein neues hearObject daraus erzeugen.
 *
 * @see <code>hearObject</code>
 * @see rcssmanual-7.07.pdf 4.2.2: client sensor protocol
 */
      time = Integer.parseInt( initErrorString.nextToken().trim() );
      String sender = initErrorString.nextToken().trim();
/** den eigenen ruf muessen wir nicht extra hoeren */
      if( !sender.equals( "self" ) ) {
        hearObject ho = new hearObject( time,
                                        sender,
                                        initErrorString.nextToken().trim()
                                      );
        this.ho = ho;
        neuesHearObject = true;
      }
    }
  } else {
// alle anderen serverDaten mit klammern und mit zeitangabe...
/** type {see|sense_body} etc */
    StringTokenizer serverTokens = new StringTokenizer( serverDaten, "()" );
    StringTokenizer typeTime = new StringTokenizer( serverTokens.nextToken(), " " );
    type = typeTime.nextToken().trim();
/** time {>=0} */
    time = Integer.parseInt( typeTime.nextToken() );
/** "see" message? */
    if( ( type.startsWith( "see" ) ) && !( neuesSeeObject ) ) {
//      System.out.println( "Kommunikation: see message: timestamp = "+time );
// see message bearbeiten
      seeObject[] so;
      String seeMessage = "";
      int seeObjectCount = 0;
      so = new seeObject[ serverTokens.countTokens() ];
      while( serverTokens.hasMoreTokens() ) {
/** in seeMessage steht jetzt alles nach dem "see <timestamp>" */
        seeMessage = serverTokens.nextToken().trim();
/**
 * in seeMessage steht jetzt alles nach dem "see <timestamp>"
 * die gesendete nachricht ohne den typ und den timestamp in die einzelnen parameter aufteilen
 * und ein neues seeObject daraus erzeugen.
 *
 * @see <code>seeObject</code>
 * @see rcssmanual-7.07.pdf 4.2.2: client sensor protocol
 */
        if( !seeMessage.equals( "" ) ) {
          String objName = seeMessage;
          StringTokenizer objParam = new StringTokenizer( serverTokens.nextToken().trim(), " " );
/** wir bekommen nur die direction des objects */
          if( objParam.countTokens() == 1 ) {
            so[ seeObjectCount++ ] = new seeObject( time, objName, Double.parseDouble( objParam.nextToken().trim() ) );
          }
/** wir bekommen distance und direction */
          if( objParam.countTokens() == 2 ) {
            so[ seeObjectCount++ ] = new seeObject( time, objName,
                                                    Double.parseDouble( objParam.nextToken().trim() ),
                                                    Double.parseDouble( objParam.nextToken().trim() ) );
          }
/** wir bekommen distance, direction, dirChange, distChange */
          if( objParam.countTokens() == 4 ) {
            so[ seeObjectCount++ ] = new seeObject( time, objName,
                                                    Double.parseDouble( objParam.nextToken().trim() ),
                                                    Double.parseDouble( objParam.nextToken().trim() ),
                                                    Double.parseDouble( objParam.nextToken().trim() ),
                                                    Double.parseDouble( objParam.nextToken().trim() ) );
          }
/** wir bekommen distance, direction, dirChange, distChange, bodyFacingDir, headFacingDir */
          if( objParam.countTokens() == 6 ) {
            so[ seeObjectCount++ ] = new seeObject( time, objName,
                                                    Double.parseDouble( objParam.nextToken().trim() ),
                                                    Double.parseDouble( objParam.nextToken().trim() ),
                                                    Double.parseDouble( objParam.nextToken().trim() ),
                                                    Double.parseDouble( objParam.nextToken().trim() ),
                                                    Double.parseDouble( objParam.nextToken().trim() ),
                                                    Double.parseDouble( objParam.nextToken().trim() ) );
            }
        }
      }
      this.so = so;
      neuesSeeObject = true;
    }
/** "sense_body" message? */
    if( ( type.startsWith( "sense_body" ) )  && !( neuesSense_bodyObject ) ) {
//      System.out.println( "Kommunikation: sense_body message: timestamp = "+time );
// sense_body message bearbeiten
// view mode + quality
      String sense_bodyMessage = serverTokens.nextToken().trim();
      String viewMode = sense_bodyMessage.substring( sense_bodyMessage.indexOf( " " ), sense_bodyMessage.lastIndexOf( " " ) ).trim();
      String viewQuality = sense_bodyMessage.substring( sense_bodyMessage.lastIndexOf( " " ) ).trim();
// stamina + effort
      do { sense_bodyMessage = serverTokens.nextToken().trim(); } while( sense_bodyMessage.equals( "" ) );
      double stamina = Double.parseDouble( sense_bodyMessage.substring( sense_bodyMessage.indexOf( " " ), sense_bodyMessage.lastIndexOf( " " ) ).trim() );
      double effort = Double.parseDouble ( sense_bodyMessage.substring( sense_bodyMessage.lastIndexOf( " " ) ).trim() );
// amountOfSpeed + directionOfSpeed
      do { sense_bodyMessage = serverTokens.nextToken().trim(); } while( sense_bodyMessage.equals( "" ) );
      double amountOfSpeed = Double.parseDouble( sense_bodyMessage.substring( sense_bodyMessage.indexOf( " " ), sense_bodyMessage.lastIndexOf( " " ) ).trim() );
      double directionOfSpeed = Double.parseDouble ( sense_bodyMessage.substring( sense_bodyMessage.lastIndexOf( " " ) ).trim() );
// headDirection
      do { sense_bodyMessage = serverTokens.nextToken().trim(); } while( sense_bodyMessage.equals( "" ) );
      double headDirection = Double.parseDouble( sense_bodyMessage.substring( sense_bodyMessage.indexOf( " " ) ).trim() );
// kickCount
      do { sense_bodyMessage = serverTokens.nextToken().trim(); } while( sense_bodyMessage.equals( "" ) );
      int kickCount = Integer.parseInt( sense_bodyMessage.substring( sense_bodyMessage.indexOf( " " ) ).trim() );
// dashCount
      do { sense_bodyMessage = serverTokens.nextToken().trim(); } while( sense_bodyMessage.equals( "" ) );
      int dashCount = Integer.parseInt( sense_bodyMessage.substring( sense_bodyMessage.indexOf( " " ) ).trim() );
// turnCount
      do { sense_bodyMessage = serverTokens.nextToken().trim(); } while( sense_bodyMessage.equals( "" ) );
      int turnCount = Integer.parseInt( sense_bodyMessage.substring( sense_bodyMessage.indexOf( " " ) ).trim() );
// sayCount
      do { sense_bodyMessage = serverTokens.nextToken().trim(); } while( sense_bodyMessage.equals( "" ) );
      int sayCount = Integer.parseInt( sense_bodyMessage.substring( sense_bodyMessage.indexOf( " " ) ).trim() );
// turnNeckCount
      do { sense_bodyMessage = serverTokens.nextToken().trim(); } while( sense_bodyMessage.equals( "" ) );
      int turnNeckCount = Integer.parseInt( sense_bodyMessage.substring( sense_bodyMessage.indexOf( " " ) ).trim() );
// catchCount
      do { sense_bodyMessage = serverTokens.nextToken().trim(); } while( sense_bodyMessage.equals( "" ) );
      int catchCount = Integer.parseInt( sense_bodyMessage.substring( sense_bodyMessage.indexOf( " " ) ).trim() );
// moveCount
      do { sense_bodyMessage = serverTokens.nextToken().trim(); } while( sense_bodyMessage.equals( "" ) );
      int moveCount = Integer.parseInt( sense_bodyMessage.substring( sense_bodyMessage.indexOf( " " ) ).trim() );
// changeViewCount
      do { sense_bodyMessage = serverTokens.nextToken().trim(); } while( sense_bodyMessage.equals( "" ) );
      int changeViewCount = Integer.parseInt( sense_bodyMessage.substring( sense_bodyMessage.indexOf( " " ) ).trim() );
//
// alle daten geparst, object erzeugen...
//
     sense_bodyObject s_bo = new sense_bodyObject( time,
                                                   viewQuality, viewMode,
                                                   stamina, effort,
                                                   amountOfSpeed, directionOfSpeed,
                                                   headDirection,
                                                   kickCount, dashCount, turnCount,
                                                   sayCount, turnNeckCount, catchCount,
                                                   moveCount, changeViewCount );
      this.s_bo = s_bo;
      neuesSense_bodyObject = true;
    }
  }
// serverDaten abgearbeitet, erstmal wieder loeschen...
  serverDaten = "";
}
// ende von bearbeiteServerDaten( String serverDaten )

}
// ende der kommunikations hauptklasse

// ende des kommunikationsmodules
/**
 * -----
 * created listening to:
 * MCM, MCM2, MCM Africa, VH1, VH1 Classics
 * RockAntenne, ClassicRock, AlternativeRock
 * ...and winamp ;)
 * -----
 */